@@ -1,5 +1,5 @@
 /* This file is part of libmspack.
- * (C) 2003-2004 Stuart Caie.
+ * (C) 2003-2010 Stuart Caie.
  *
  * The deflate method was created by Phil Katz. MSZIP is equivalent to the
  * deflate method.
@@ -15,6 +15,34 @@
 #include <system.h>
 #include <mszip.h>
 
+/* import bit-reading macros and code */
+#define BITS_TYPE struct mszipd_stream
+#define BITS_VAR zip
+#define BITS_ORDER_LSB
+#define BITS_LSB_TABLE
+#define READ_BYTES do {		\
+    READ_IF_NEEDED;		\
+    INJECT_BITS(*i_ptr++, 8);	\
+} while (0)
+#include <readbits.h>
+
+/* import huffman macros and code */
+#define TABLEBITS(tbl)      MSZIP_##tbl##_TABLEBITS
+#define MAXSYMBOLS(tbl)     MSZIP_##tbl##_MAXSYMBOLS
+#define HUFF_TABLE(tbl,idx) zip->tbl##_table[idx]
+#define HUFF_LEN(tbl,idx)   zip->tbl##_len[idx]
+#define HUFF_ERROR          return INF_ERR_HUFFSYM
+#include <readhuff.h>
+
+#define FLUSH_IF_NEEDED do {				\
+    if (zip->window_posn == MSZIP_FRAME_SIZE) {		\
+	if (zip->flush_window(zip, MSZIP_FRAME_SIZE)) {	\
+	    return INF_ERR_FLUSH;			\
+	}						\
+	zip->window_posn = 0;				\
+    }							\
+} while (0)
+
 /* match lengths for literal codes 257.. 285 */
 static const unsigned short lit_lengths[29] = {
   3, 4, 5, 6, 7, 8, 9, 10, 11, 13, 15, 17, 19, 23, 27,
@@ -44,59 +72,6 @@ static const unsigned char bitlen_order[19] = {
   16, 17, 18, 0, 8, 7, 9, 6, 10, 5, 11, 4, 12, 3, 13, 2, 14, 1, 15
 };
 
-/* ANDing with bit_mask[n] masks the lower n bits */
-static const unsigned short bit_mask[17] = {
- 0x0000, 0x0001, 0x0003, 0x0007, 0x000f, 0x001f, 0x003f, 0x007f, 0x00ff,
- 0x01ff, 0x03ff, 0x07ff, 0x0fff, 0x1fff, 0x3fff, 0x7fff, 0xffff
-};
-
-#define STORE_BITS do {                                                 \
-  zip->i_ptr      = i_ptr;                                              \
-  zip->i_end      = i_end;                                              \
-  zip->bit_buffer = bit_buffer;                                         \
-  zip->bits_left  = bits_left;                                          \
-} while (0)
-
-#define RESTORE_BITS do {                                               \
-  i_ptr      = zip->i_ptr;                                              \
-  i_end      = zip->i_end;                                              \
-  bit_buffer = zip->bit_buffer;                                         \
-  bits_left  = zip->bits_left;                                          \
-} while (0)
-
-#define ENSURE_BITS(nbits) do {                                         \
-  while (bits_left < (nbits)) {                                         \
-    if (i_ptr >= i_end) {                                               \
-      if (zipd_read_input(zip)) return zip->error;                      \
-      i_ptr = zip->i_ptr;                                               \
-      i_end = zip->i_end;                                               \
-    }                                                                   \
-    bit_buffer |= *i_ptr++ << bits_left; bits_left  += 8;               \
-  }                                                                     \
-} while (0)
-
-#define PEEK_BITS(nbits)   (bit_buffer & ((1<<(nbits))-1))
-#define PEEK_BITS_T(nbits) (bit_buffer & bit_mask[(nbits)])
-
-#define REMOVE_BITS(nbits) ((bit_buffer >>= (nbits)), (bits_left -= (nbits)))
-
-#define READ_BITS(val, nbits) do {                                      \
-  ENSURE_BITS(nbits); (val) = PEEK_BITS(nbits); REMOVE_BITS(nbits);     \
-} while (0)
-
-#define READ_BITS_T(val, nbits) do {                                    \
-  ENSURE_BITS(nbits); (val) = PEEK_BITS_T(nbits); REMOVE_BITS(nbits);   \
-} while (0)
-
-static int zipd_read_input(struct mszipd_stream *zip) {
-  int read = zip->sys->read(zip->input, &zip->inbuf[0], (int)zip->inbuf_size);
-  if (read < 0) return zip->error = MSPACK_ERR_READ;
-  zip->i_ptr = &zip->inbuf[0];
-  zip->i_end = &zip->inbuf[read];
-
-  return MSPACK_ERR_OK;
-}
-
 /* inflate() error codes */
 #define INF_ERR_BLOCKTYPE   (-1)  /* unknown block type                      */
 #define INF_ERR_COMPLEMENT  (-2)  /* block size complement mismatch          */
@@ -113,129 +88,6 @@ static int zipd_read_input(struct mszipd_stream *zip) {
 #define INF_ERR_DISTANCE    (-13) /* somehow, distance is beyond 32k         */
 #define INF_ERR_HUFFSYM     (-14) /* out of bits decoding huffman symbol     */
 
-/* make_decode_table(nsyms, nbits, length[], table[])
- *
- * This function was coded by David Tritscher. It builds a fast huffman
- * decoding table out of just a canonical huffman code lengths table.
- *
- * NOTE: this is NOT identical to the make_decode_table() in lzxd.c. This
- * one reverses the quick-lookup bit pattern. Bits are read MSB to LSB in LZX,
- * but LSB to MSB in MSZIP.
- *
- * nsyms  = total number of symbols in this huffman tree.
- * nbits  = any symbols with a code length of nbits or less can be decoded
- *          in one lookup of the table.
- * length = A table to get code lengths from [0 to nsyms-1]
- * table  = The table to fill up with decoded symbols and pointers.
- *
- * Returns 0 for OK or 1 for error
- */
-static int make_decode_table(unsigned int nsyms, unsigned int nbits,
-			     unsigned char *length, unsigned short *table)
-{
-  register unsigned int leaf, reverse, fill;
-  register unsigned short sym, next_sym;
-  register unsigned char bit_num;
-  unsigned int pos         = 0; /* the current position in the decode table */
-  unsigned int table_mask  = 1 << nbits;
-  unsigned int bit_mask    = table_mask >> 1; /* don't do 0 length codes */
-
-  /* fill entries for codes short enough for a direct mapping */
-  for (bit_num = 1; bit_num <= nbits; bit_num++) {
-    for (sym = 0; sym < nsyms; sym++) {
-      if (length[sym] != bit_num) continue;
-
-      /* reverse the significant bits */
-      fill = length[sym]; reverse = pos >> (nbits - fill); leaf = 0;
-      do {leaf <<= 1; leaf |= reverse & 1; reverse >>= 1;} while (--fill);
-
-      if((pos += bit_mask) > table_mask) return 1; /* table overrun */
-
-      /* fill all possible lookups of this symbol with the symbol itself */
-      fill = bit_mask; next_sym = 1 << bit_num;
-      do { table[leaf] = sym; leaf += next_sym; } while (--fill);
-    }
-    bit_mask >>= 1;
-  }
-
-  /* exit with success if table is now complete */
-  if (pos == table_mask) return 0;
-
-  /* mark all remaining table entries as unused */
-  for (sym = pos; sym < table_mask; sym++) {
-    reverse = sym; leaf = 0; fill = nbits;
-    do { leaf <<= 1; leaf |= reverse & 1; reverse >>= 1; } while (--fill);
-    table[leaf] = 0xFFFF;
-  }
-
-  /* where should the longer codes be allocated from? */
-  next_sym = ((table_mask >> 1) < nsyms) ? nsyms : (table_mask >> 1);
-
-  /* give ourselves room for codes to grow by up to 16 more bits.
-   * codes now start at bit nbits+16 and end at (nbits+16-codelength) */
-  pos <<= 16;
-  table_mask <<= 16;
-  bit_mask = 1 << 15;
-
-  for (bit_num = nbits+1; bit_num <= MSZIP_MAX_HUFFBITS; bit_num++) {
-    for (sym = 0; sym < nsyms; sym++) {
-      if (length[sym] != bit_num) continue;
-
-      /* leaf = the first nbits of the code, reversed */
-      reverse = pos >> 16; leaf = 0; fill = nbits;
-      do {leaf <<= 1; leaf |= reverse & 1; reverse >>= 1;} while (--fill);
-
-      for (fill = 0; fill < (bit_num - nbits); fill++) {
-	/* if this path hasn't been taken yet, 'allocate' two entries */
-	if (table[leaf] == 0xFFFF) {
-	  table[(next_sym << 1)     ] = 0xFFFF;
-	  table[(next_sym << 1) + 1 ] = 0xFFFF;
-	  table[leaf] = next_sym++;
-	}
-	/* follow the path and select either left or right for next bit */
-	leaf = (table[leaf] << 1) | ((pos >> (15 - fill)) & 1);
-      }
-      table[leaf] = sym;
-
-      if ((pos += bit_mask) > table_mask) return 1; /* table overflow */
-    }
-    bit_mask >>= 1;
-  }
-
-  /* full table? */
-  return (pos != table_mask) ? 1 : 0;
-}
-
-/* READ_HUFFSYM(tablename, var) decodes one huffman symbol from the
- * bitstream using the stated table and puts it in var.
- */
-#define READ_HUFFSYM(tbl, var) do {                                     \
-  /* huffman symbols can be up to 16 bits long */                       \
-  ENSURE_BITS(MSZIP_MAX_HUFFBITS);                                      \
-  /* immediate table lookup of [tablebits] bits of the code */          \
-  sym = zip->tbl##_table[PEEK_BITS(MSZIP_##tbl##_TABLEBITS)];		\
-  /* is the symbol is longer than [tablebits] bits? (i=node index) */   \
-  if (sym >= MSZIP_##tbl##_MAXSYMBOLS) {                                \
-    /* decode remaining bits by tree traversal */                       \
-    i = MSZIP_##tbl##_TABLEBITS - 1;					\
-    do {                                                                \
-      /* check next bit. error if we run out of bits before decode */	\
-      if (i++ > MSZIP_MAX_HUFFBITS) {					\
-        D(("out of bits in huffman decode"))                            \
-        return INF_ERR_HUFFSYM;                                         \
-      }                                                                 \
-      /* double node index and add 0 (left branch) or 1 (right) */	\
-      sym = zip->tbl##_table[(sym << 1) | ((bit_buffer >> i) & 1)];	\
-      /* while we are still in node indicies, not decoded symbols */    \
-    } while (sym >= MSZIP_##tbl##_MAXSYMBOLS);                          \
-  }                                                                     \
-  /* result */                                                          \
-  (var) = sym;                                                          \
-  /* look up the code length of that symbol and discard those bits */   \
-  i = zip->tbl##_len[sym];                                              \
-  REMOVE_BITS(i);                                                       \
-} while (0)
-
 static int zip_read_lens(struct mszipd_stream *zip) {
   /* for the bit buffer and huffman decoding */
   register unsigned int bit_buffer;
@@ -336,11 +188,7 @@ static int inflate(struct mszipd_stream *zip) {
       }
       if (bits_left != 0) return INF_ERR_BITBUF;
       while (i < 4) {
-	if (i_ptr >= i_end) {
-	  if (zipd_read_input(zip)) return zip->error;
-	  i_ptr = zip->i_ptr;
-	  i_end = zip->i_end;
-	}
+	READ_IF_NEEDED;
 	lens_buf[i++] = *i_ptr++;
       }
 
@@ -351,11 +199,7 @@ static int inflate(struct mszipd_stream *zip) {
 
       /* read and copy the uncompressed data into the window */
       while (length > 0) {
-	if (i_ptr >= i_end) {
-	  if (zipd_read_input(zip)) return zip->error;
-	  i_ptr = zip->i_ptr;
-	  i_end = zip->i_end;
-	}
+	READ_IF_NEEDED;
 
 	this_run = length;
 	if (this_run > (unsigned int)(i_end - i_ptr)) this_run = i_end - i_ptr;
@@ -366,16 +210,12 @@ static int inflate(struct mszipd_stream *zip) {
 	zip->window_posn += this_run;
 	i_ptr    += this_run;
 	length   -= this_run;
-
-	if (zip->window_posn == MSZIP_FRAME_SIZE) {
-	  if (zip->flush_window(zip, MSZIP_FRAME_SIZE)) return INF_ERR_FLUSH;
-	  zip->window_posn = 0;
-	}
+	FLUSH_IF_NEEDED;
       }
     }
     else if ((block_type == 1) || (block_type == 2)) {
       /* Huffman-compressed LZ77 block */
-      unsigned int window_posn, match_posn, code;
+      unsigned int match_posn, code;
 
       if (block_type == 1) {
 	/* block with fixed Huffman codes */
@@ -408,23 +248,19 @@ static int inflate(struct mszipd_stream *zip) {
       }
 
       /* decode forever until end of block code */
-      window_posn = zip->window_posn;
-      while (1) {
+      for (;;) {
 	READ_HUFFSYM(LITERAL, code);
 	if (code < 256) {
-	  zip->window[window_posn++] = (unsigned char) code;
-	  if (window_posn == MSZIP_FRAME_SIZE) {
-	    if (zip->flush_window(zip, MSZIP_FRAME_SIZE)) return INF_ERR_FLUSH;
-	    window_posn = 0;
-	  }
+	  zip->window[zip->window_posn++] = (unsigned char) code;
+	  FLUSH_IF_NEEDED;
 	}
 	else if (code == 256) {
 	  /* END OF BLOCK CODE: loop break point */
 	  break;
 	}
 	else {
-	  code -= 257;
-	  if (code > 29) return INF_ERR_LITCODE;
+	  code -= 257; /* codes 257-285 are matches */
+	  if (code >= 29) return INF_ERR_LITCODE; /* codes 286-287 are illegal */
 	  READ_BITS_T(length, lit_extrabits[code]);
 	  length += lit_lengths[code];
 
@@ -436,21 +272,16 @@ static int inflate(struct mszipd_stream *zip) {
 	  /* match position is window position minus distance. If distance
 	   * is more than window position numerically, it must 'wrap
 	   * around' the frame size. */ 
-	  match_posn = ((distance > window_posn) ? MSZIP_FRAME_SIZE : 0)
-	    + window_posn - distance;
+	  match_posn = ((distance > zip->window_posn) ? MSZIP_FRAME_SIZE : 0)
+	    + zip->window_posn - distance;
 
 	  /* copy match */
 	  if (length < 12) {
 	    /* short match, use slower loop but no loop setup code */
 	    while (length--) {
-	      zip->window[window_posn++] = zip->window[match_posn++];
+	      zip->window[zip->window_posn++] = zip->window[match_posn++];
 	      match_posn &= MSZIP_FRAME_SIZE - 1;
-
-	      if (window_posn == MSZIP_FRAME_SIZE) {
-		if (zip->flush_window(zip, MSZIP_FRAME_SIZE))
-		  return INF_ERR_FLUSH;
-		window_posn = 0;
-	      }
+	      FLUSH_IF_NEEDED;
 	    }
 	  }
 	  else {
@@ -460,28 +291,21 @@ static int inflate(struct mszipd_stream *zip) {
 	      this_run = length;
 	      if ((match_posn + this_run) > MSZIP_FRAME_SIZE)
 		this_run = MSZIP_FRAME_SIZE - match_posn;
-	      if ((window_posn + this_run) > MSZIP_FRAME_SIZE)
-		this_run = MSZIP_FRAME_SIZE - window_posn;
+	      if ((zip->window_posn + this_run) > MSZIP_FRAME_SIZE)
+		this_run = MSZIP_FRAME_SIZE - zip->window_posn;
 
-	      rundest = &zip->window[window_posn]; window_posn += this_run;
+	      rundest = &zip->window[zip->window_posn]; zip->window_posn += this_run;
 	      runsrc  = &zip->window[match_posn];  match_posn  += this_run;
 	      length -= this_run;
 	      while (this_run--) *rundest++ = *runsrc++;
-
-	      /* flush if necessary */
-	      if (window_posn == MSZIP_FRAME_SIZE) {
-		if (zip->flush_window(zip, MSZIP_FRAME_SIZE))
-		  return INF_ERR_FLUSH;
-		window_posn = 0;
-	      }
 	      if (match_posn == MSZIP_FRAME_SIZE) match_posn = 0;
+	      FLUSH_IF_NEEDED;
 	    } while (length > 0);
 	  }
 
 	} /* else (code >= 257) */
 
-      } /* while (forever) -- break point at 'code == 256' */
-      zip->window_posn = window_posn;
+      } /* for(;;) -- break point at 'code == 256' */
     }
     else {
       /* block_type == 3 -- bad block type */
@@ -546,6 +370,7 @@ struct mszipd_stream *mszipd_init(struct mspack_system *system,
   zip->input           = input;
   zip->output          = output;
   zip->inbuf_size      = input_buffer_size;
+  zip->input_end       = 0;
   zip->error           = MSPACK_ERR_OK;
   zip->repair_mode     = repair_mode;
   zip->flush_window    = &mszipd_flush_window;
@@ -600,8 +425,12 @@ int mszipd_decompress(struct mszipd_stream *zip, off_t out_bytes) {
     zip->bytes_output = 0;
     STORE_BITS;
     if ((error = inflate(zip))) {
-      D(("inflate error %d", i))
+      D(("inflate error %d", error))
       if (zip->repair_mode) {
+	/* recover partially-inflated buffers */
+	if (zip->bytes_output == 0 && zip->window_posn > 0) {
+	  zip->flush_window(zip, zip->window_posn);
+	}
 	zip->sys->message(NULL, "MSZIP error, %u bytes of data lost.",
 			  MSZIP_FRAME_SIZE - zip->bytes_output);
 	for (i = zip->bytes_output; i < MSZIP_FRAME_SIZE; i++) {
